@@ -19,7 +19,7 @@
  * Portions created by the Initial Developer are Copyright (C) 2008
  * the Initial Developer. All Rights Reserved.
  *
- * Contributor(s): Jason Orendorff
+ * Contributor(s): Jesse Ruderman
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -35,7 +35,7 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-var gTestfile = 'regress-452498-052-a.js';
+var gTestfile = 'regress-452498-075.js';
 //-----------------------------------------------------------------------------
 var BUGNUMBER = 452498;
 var summary = 'TM: upvar2 regression tests';
@@ -53,20 +53,13 @@ function test()
   printBugNumber(BUGNUMBER);
   printStatus (summary);
 
-// ------- Comment #52 From Jason Orendorff
+// ------- Comment #75 From Jesse Ruderman
 
+  (function p(){ p = 3; });
+  (function p(){ p = 3; return p; })()
+// Assertion failure: regs.sp == StackBase(fp), at ../jsinterp.cpp:2980
 
-// Assertion failure: pn_arity == PN_FUNC || pn_arity == PN_NAME, at ../jsparse.h:444
-// Here the function node has been morphed into a JSOP_TRUE node, but we're in
-// FindFunArgs poking it anyway.
-  if (typeof timeout == 'function')
-  {
-    expectExitCode(6);
-    timeout(3);
-    while(function(){});
-  }
-
-  reportCompare(expect, actual, summary);
+    reportCompare(expect, actual, summary);
 
   exitFunc ('test');
 }
